/*

This pam module uses the Heimdal libraries to generate a KRB5 TGT out of
a globus proxy certificate. It is up to other PAM modules (like pam_krb5)
to eventually generate an AFS token also.

*/

#define SERVICENAME "pam_gridpxy2krb5"

#define PAM_SM_AUTH
#define PAM_SM_ACCOUNT
#define PAM_SM_SESSION
#define PAM_SM_PASSWORD

#define MAX_STRLEN 128

#ifndef PAM_EXTERN
#define PAM_EXTERN extern
#endif

#ifdef SOLARIS
#define PAM_CONST
#else
#define PAM_CONST const
#endif

#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>

#include <pwd.h>
#include <security/pam_appl.h>
#include <sys/types.h>
#include <syslog.h>
#include <sys/syslog.h>

#include <krb5.h>

#ifndef KRB5_SUCCESS
#define KRB5_SUCCESS 0
#endif

/* predeclare some functions */

void do_log(int priority, const char *, ...);
PAM_EXTERN int pam_sm_open_session(pam_handle_t *pamh,
    int flags, int argc, const char **argv);

/* implemenent all pam methods the pam stack contains */

PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh,
    int flags, int argc, const char **argv)
{
    return PAM_SUCCESS;
}

PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh,
    int flags, int argc, const char **argv)
{
    if ( flags & PAM_REINITIALIZE_CRED )
	return pam_sm_open_session(pamh, flags, argc, argv);
    return PAM_SUCCESS;
}

PAM_EXTERN int pam_sm_acct_mgmt(pam_handle_t *pamh,
    int flags, int argc, const char **argv)
{
    return PAM_SUCCESS;
}

PAM_EXTERN int pam_sm_chauthtok(pam_handle_t *pamh,
    int flags, int argc, const char **argv)
{
    return PAM_SUCCESS;
}

PAM_EXTERN int pam_sm_open_session(pam_handle_t *pamh,
    int flags, int argc, const char **argv)
{
    const char *env_x509 = NULL, *user = NULL;
    char krb5_cc[MAX_STRLEN] = "",
	 krb5_x509[MAX_STRLEN] = "",
	 pam_env_krb5[MAX_STRLEN] = "",
	 env_krb5[MAX_STRLEN] = "";
    struct passwd *passwd_entry = NULL;

    krb5_context context = NULL;
    krb5_principal principal = NULL;
    krb5_ccache ccache = NULL;
    krb5_creds cred;
    krb5_get_init_creds_opt *krb5_opts = NULL;

    int rc = PAM_SUCCESS, fd = 0;

    if ( ! (env_x509 = pam_getenv(pamh, "X509_USER_PROXY")) ) {
	do_log(LOG_DEBUG, "env X509_USER_PROXY not set...");
	return PAM_SUCCESS;
    }
    snprintf(krb5_x509, MAX_STRLEN, "FILE:%s", env_x509);
    do_log(LOG_DEBUG, "using proxy: %s", krb5_x509);

    if ( (pam_get_user(pamh, (PAM_CONST char **) &user, NULL) != PAM_SUCCESS) ||
	 (user == NULL) ) {
	do_log(LOG_ERR, "couldn't determine user name from PAM");
	return PAM_SYSTEM_ERR;
    }
    if ( ( passwd_entry = getpwnam(user) ) == NULL ) {
	do_log(LOG_ERR, "couldn't determine user %s's passwd entry", user);
	return PAM_SYSTEM_ERR;
    }

    snprintf(krb5_cc, MAX_STRLEN, "/tmp/krb5cc_%d_XXXXXX",
	passwd_entry->pw_uid);
    if ( fd = mkstemp(krb5_cc) ) {
	close(fd);
    } else {
	do_log(LOG_ERR, "failed to generate name for krb5 credentials cache");
	return PAM_SYSTEM_ERR;
    }
    snprintf(env_krb5, MAX_STRLEN, "FILE:%s", krb5_cc);
    snprintf(pam_env_krb5, MAX_STRLEN, "KRB5CCNAME=%s", env_krb5);
    do_log(LOG_DEBUG, "setting %s", pam_env_krb5);
    pam_putenv(pamh, pam_env_krb5);

    if ( krb5_init_context(&context) != KRB5_SUCCESS )
	return PAM_SYSTEM_ERR;

    memset(&cred, 0, sizeof(cred));

    if ( (rc = krb5_parse_name(context, user, &principal)) != KRB5_SUCCESS ) {
	do_log(LOG_ERR, "error building user %s's principal name: %s",
	    user, error_message(rc));
	rc = PAM_SYSTEM_ERR;
	goto out2;
    }

    if ( (rc = krb5_cc_resolve(context, env_krb5, &ccache)) != KRB5_SUCCESS ) {
	do_log(LOG_ERR, "error resolving credentials cache file: %s",
	    error_message(rc));
	rc = PAM_SYSTEM_ERR;
	goto out2;
    }

    if ( (rc = krb5_get_init_creds_opt_alloc(context, &krb5_opts))
	  != KRB5_SUCCESS) {
	do_log(LOG_ERR, "failed to krb5_get_init_creds_opt_alloc: %s",
	    error_message(rc));
	rc = PAM_SYSTEM_ERR;
	goto out2;
    }
    krb5_get_init_creds_opt_set_default_flags(context, SERVICENAME,
	krb5_principal_get_realm(context, principal), krb5_opts);

    rc = krb5_get_init_creds_opt_set_pkinit(context, krb5_opts, principal,
	krb5_x509, NULL, NULL, NULL, 0, NULL, NULL, NULL);
    if ( rc != KRB5_SUCCESS ) {
	do_log(LOG_ERR, "failed to set krb5 options: %s",
	    error_message(rc));
	rc = PAM_SYSTEM_ERR;
	goto out2;
    }

    rc = krb5_get_init_creds_password(context, &cred, principal,
	"", NULL, NULL, 0, NULL, krb5_opts);
    if ( rc != KRB5_SUCCESS ) {
	do_log(LOG_ERR, "failed to generate TGT out of grid proxy: %s",
	    error_message(rc));
	rc = PAM_SYSTEM_ERR;
	goto out1;
    }

    if ( (rc = krb5_cc_initialize(context, ccache, cred.client))
	    != KRB5_SUCCESS ) {
	do_log(LOG_ERR, "failed to initialize credentials: %s",
	    error_message(rc));
	rc = PAM_SYSTEM_ERR;
	goto out1;
    }

    /* store cc to file */
    if ( (rc = krb5_cc_store_cred(context, ccache, &cred)) != KRB5_SUCCESS ) {
	do_log(LOG_ERR, "failed to store credentials: %s", error_message(rc));
	rc = PAM_SYSTEM_ERR;
	goto out1;
    }
    krb5_cc_close(context, ccache);
    chown(krb5_cc, passwd_entry->pw_uid, passwd_entry->pw_gid);

out1:
    krb5_free_cred_contents(context, &cred);
out2:
    if ( principal != NULL ) krb5_free_principal(context, principal);
    if ( context != NULL ) krb5_free_context(context);

    return rc;
}

PAM_EXTERN int pam_sm_close_session(pam_handle_t *pamh,
    int flags, int argc, const char **argv)
{
    return PAM_SUCCESS;
}

void do_log(int priority, const char *msg, ...)
{
    char buf[512];
    va_list plist;
    va_start(plist, msg);
    vsnprintf(buf, sizeof(buf), msg, plist);
    va_end(plist);
    openlog(SERVICENAME, LOG_PID | LOG_CONS | LOG_NOWAIT, LOG_AUTH);
    syslog(priority, buf);
    closelog();
}

